/*----------------------------------------------------------------------
  File    : wbcview.c
  Contents: Bayes classifier visualization program for Windows
  Author  : Christian Borgelt
  History : 2000.01.28 file created from winskel.c
            2000.01.30 adapted to data files without a class column
            2003.04.13 clustering result visualization added
            2004.03.19 bug in command line evaluation fixed
            2004.04.27 redrawing simplified with function bvn_dist
            2004.05.15 changable data marker size added
            2004.04.24 bug in command line evaluation fixed
            2007.02.18 bug in command line evaluation fixed
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <windows.h>
#include <commdlg.h>
#include "wbcview.rh"
#include "wshades.h"
#include "bcload.h"

#if ATTID_MAX > INT_MAX
#error "wbcview requires ATTID_MAX <= INT_MAX"
#endif
#if VALID_MAX > INT_MAX
#error "wbcview requires VALID_MAX <= INT_MAX"
#endif
#if DIMID_MAX > INT_MAX
#error "wbcview requires DIMID_MAX <= INT_MAX"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "WBCView"   /* main window title */

#ifndef PATH_MAX
#define PATH_MAX    1024        /* max. len. of a path name */
#endif

#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI        3.14159265358979323846  /* \pi */
#endif

/* --- abbreviations --- */
#define HINST       HINSTANCE
#define MKIRSC      MAKEINTRESOURCE
#define SDIMSG      SendDlgItemMessage
#define WINSTYLE    WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN

/* --- extensions --- */
#define MINXEXT     160         /* minimal x-extension of window */
#define MINYEXT     120         /* minimal y-extension of window */
#define INITXEXT    320         /* initial horizontal extension */
#define INITYEXT    320         /* initial vertical   extension */

/* --- edit box functions --- */
#define edt_set(h,i,s)    SetDlgItemText(h, i, s)
#define edt_get(h,i,s,n)  GetDlgItemText(h, i, s, n)
#define int_set(h,i,n)    SetDlgItemInt(h, i, n, 0)
#define int_get(h,i)      GetDlgItemInt(h, i, NULL, 0)

/* --- check box functions --- */
#define chk_set(h,i,s)    CheckDlgButton(h, i, s)
#define chk_get(h,i)      IsDlgButtonChecked(h, i)

/* --- combo box functions --- */
#define cmb_add(h,i,s)    SDIMSG(h, i, CB_ADDSTRING,    0, (LPARAM)(s))
#define cmb_rem(h,i,k)    SDIMSG(h, i, CB_DELETESTRING, k, 0)
#define cmb_set(h,i,k)    SDIMSG(h, i, CB_SETCURSEL,    k, 0)
#define cmb_get(h,i)      SDIMSG(h, i, CB_GETCURSEL,    0, 0)

/* --- marker flags --- */
#define MRK_DATA    0x01        /* mark data points */
#define MRK_COLOR   0x02        /* use color on colored shades */
#define MRK_CENTER  0x10        /* mark centers of distributions */
#define MRK_ELL1S   0x20        /* draw 1 sigma ellipses */
#define MRK_ELL2S   0x40        /* draw 2 sigma ellipses */
#define MRK_ELL3S   0x80        /* draw 3 sigma ellipses */

/* --- distribution types --- */
#define DT_PROB        0        /* probabilistic */
#define DT_FUZZY       1        /* fuzzy (probabilistic) */
#define DT_POSS        2        /* fuzzy (possibilistic) */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_COLOR      (-9)       /* cannot allocate colors */
#define E_UNKNOWN   (-21)       /* unkown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
/* --- error messages --- */
static const char *errmsgs[] ={ /* error message texts */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file:\n%s",
  /* E_FREAD    -3 */  "read error on file:\n%s",
  /* E_FWRITE   -4 */  "write error on file:\n%s",
  /* E_PARSE    -5 */  "parse error on file:\n%s",
  /* E_CREATE   -6 */  "cannot create classifier",
  /* E_ATTCNT   -7 */  "no numeric attribute",
  /* E_DATA     -8 */  "a classifier must be loaded first",
  /* E_COLOR    -9 */  "cannot allocate colors",
  /*    -11 to -15 */  NULL, NULL, NULL, NULL, NULL,
  /* E_VALUE   -16 */  "file %s, record %d:\n"
                         "invalid value %s in field %d",
  /* E_FLDCNT  -17 */  "file %s, record %d:\n"
                         "%s%d field(s) instead of %d",
  /* E_EMPFLD  -18 */  "file %s, record %d:\n"
                         "empty name%s in field %d",
  /* E_DUPFLD  -19 */  "file %s, record %d:\n"
                         "duplicate field name %s",
  /* E_MISFLD  -20 */  "file %s, record %d:\n"
                         "missing field %s",
  /* E_UNKNOWN -21 */  "unknown error" };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static HINST   hinst    = NULL; /* program instance handle */
static HWND    hmain    = NULL; /* handle of main window */
static char    buf[PATH_MAX +1024];  /* buffer for various purposes */

/* --- color and marker information --- */
static SHADES  *shades  = NULL; /* color shades for normal dists. */
static SHADES  *datcols = NULL; /* colors for data points */
static HBRUSH  *brushes = NULL; /* colored brushes for data points */
static HBRUSH  br_grey  = NULL; /* grey default brush */
static PIXEL   grey     = RGB(128, 128, 128);
static HPEN    p_white  = NULL; /* white default pen */
static HPEN    p_black  = NULL; /* black default pen */
static int     mrkflgs  = 0x73; /* flags for markers */
static int     mrksize  = 5;    /* size of data markers */

static struct {                 /* --- color information --- */
  int    dist;                  /* distribution type */
  int    shdbase;               /* shade base */
  int    shdcnt;                /* number of shades */
  int    colcnt;                /* number of colors */
  double coloff;                /* color offset (in degrees) */
  double ndwgt;                 /* weight of distribution */
  double cowgt;                 /* complement of this weight */
} colinfo = { DT_PROB, SHD_WHITE, 32, 1, 0.0, 0.8, 0.2 };

/*----------------------------------------------------------------------
  Color Functions
----------------------------------------------------------------------*/

static void col_delete (void)
{                               /* --- delete color shades */
  int i;                        /* loop variable */

  if (datcols) { shd_delete(datcols); datcols = NULL; }
  if (shades)  { shd_delete(shades);  shades  = NULL; }
  if (brushes) {
    for (i = colinfo.colcnt; --i >= 0; )
      DeleteObject(brushes[i]);
    free(brushes); brushes = NULL;
  }                             /* delete all shades/colors */
}  /* col_delete() */

/*--------------------------------------------------------------------*/

static int col_create (void)
{                               /* --- create color shades */
  int clscnt;                   /* number of classes/clusters */

  if (!attset) return 0;        /* check for an existing classifier */
  col_delete();                 /* delete existing shades/colors */
  if      (nbc) clscnt = nbc_clscnt(nbc);
  else if (fbc) clscnt = fbc_clscnt(fbc);
  else if (lvq) clscnt = lvq_cnt(lvq);
  else          clscnt = cls_clscnt(clset);
  if (colinfo.colcnt > 0)       /* get the number of classes/clusters */
    colinfo.colcnt = clscnt;    /* and note the number of colors */
  shades = shd_create(colinfo.coloff,  colinfo.colcnt,
                      colinfo.shdbase, colinfo.shdcnt);
  if (!shades) return -1;       /* create color shades for distribs. */
  datcols = shd_create(colinfo.coloff, clscnt, SHD_BLACK, 1);
  if (!datcols) { col_delete(); return -1; }
  brushes = (HBRUSH*)malloc(clscnt *sizeof(HBRUSH));
  if (!brushes) { col_delete(); return -1; }
  while (--clscnt >= 0)         /* get colors for data points */
    brushes[clscnt] = CreateSolidBrush(shd_pixel(datcols, clscnt, 0));
  return 0;                     /* return 'ok' */
}  /* col_create() */

/*----------------------------------------------------------------------
  Error Reporting Function
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */
  int        n;                 /* length of the formatted message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs[-code];         /* get error message data */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  va_start(args, code);         /* get list of variable arguments */
  vsprintf(buf, msg, args);     /* get and format message */
  va_end(args);                 /* end variable argument evaluation */
  n = (int)strlen(buf) +1;      /* load alert box title */
  LoadString(hinst, CS_ALERT, buf +n, sizeof(buf) -n);
  MessageBox(hmain, buf, buf+n, MB_OK|MB_ICONERROR);
}  /* error() */                /* show alert box */

/*----------------------------------------------------------------------
  Drawing Function
----------------------------------------------------------------------*/

static void redraw (HDC hdc)
{                               /* --- redraw window contents */
  int     i, k;                 /* loop variable, buffer */
  int     s, o;                 /* size and offset for drawing */
  int     clscnt;               /* number of classes/clusters */
  int     colcnt;               /* number of colors */
  RECT    rc;                   /* clipping rectangle */
  int     x,  y;                /* pixel coordinates */
  double  rx, ry;               /* real  coordinates */
  double  dx, dy;               /* standard deviations */
  double  xoff, yoff;           /* offsets of window borders */
  double  xscl, yscl;           /* scaling factors */
  double  val, sum, max;        /* function value, sum and maximum */
  PIXEL   mrk;                  /* color value for markers */
  int     col, shd, shdcnt;     /* color, shade, and number of shades */
  HBRUSH  brush;                /* brush for drawing data points */
  double  a, b;                 /* transformation constants */
  int     xid, yid, trgid;      /* attribute identifiers */
  int     xtype, ytype;         /* attribute types */
  TUPLE   *tpl;                 /* to traverse the data tuples */
  float   f;                    /* temporary buffer */
  int     cmin, cmax;           /* minimal and maximal coordinate */
  double  t0, t1, t2;           /* temporary buffers */

  GetClientRect(hmain, &rc);    /* get the window's client rectangle */
  if (!bvnorm || !attset || !shades) {
    FillRect(hdc, &rc, GetStockObject(WHITE_BRUSH)); return; }

  /* --- initialize normal distributions --- */
  clscnt = bc_bvnorm();         /* init. the normal distributions */
  if (clscnt < 0) return;       /* and check for success */

  /* --- draw rectangle contents --- */
  xoff = ranges[selatt.h_att].min;  /* get coordinate offsets and */
  yoff = ranges[selatt.v_att].max;  /* compute the scaling factors */
  xscl = (ranges[selatt.h_att].max -xoff) /(rc.right -rc.left);
  yscl = (yoff -ranges[selatt.v_att].min) /(rc.bottom -rc.top);
  a    = 1.0/clscnt;            /* compute transformation constants */
  b    = (clscnt > 1) ? clscnt/(clscnt-1.0) : 1;
  colcnt = shd_colcnt(shades);  /* get the number of colors */
  shdcnt = shd_shdcnt(shades);  /* and the number of shades */
  GetClipBox(hdc, &rc);         /* get the rectangle to redraw */
  for (y = rc.top; y <= rc.bottom; y++) {
    for (x = rc.left; x <= rc.right; x++) {
      rx = xoff +x *xscl;       /* traverse the points and */
      ry = yoff -y *yscl;       /* compute their real coordinates */
      max = sum = 0; col = 0;   /* traverse the classes */
      if (colinfo.dist >= DT_FUZZY) {  /* if fuzzy memberships */
        for (i = clscnt; --i >= 0; ) {
          val  = bvn_dist(bvnorm +i, rx, ry);
          sum += val = (val > 0) ? 1/val : 1e12;
          if (val > max) { max = val; col = i; }
        }                       /* compute Mahalanobis distance */
        if (colinfo.dist == DT_POSS) { /* if possibilistic coloring */
          val = max /(1 +max);} /* compute relative intensity */
        else {                  /* if probabilistic fuzzy coloring */
          val = (sum > 0) ? (colinfo.ndwgt *(max/sum)
                          +  colinfo.cowgt *(max/sum -a) *b) : 0;
        } }                     /* compute relative intensity */
      else {                    /* if probabilistic coloring */
        for (i = clscnt; --i >= 0; ) {
          sum += val = bvn_eval(bvnorm +i, rx, ry);
          if (val > max) { max = val; col = i; }
        }                       /* compute class probabilities */
        val = (sum > 0) ? (colinfo.ndwgt *(max/bvn_max(bvnorm +col))
                         + colinfo.cowgt *(max/sum -a) *b) : 0;
      }                         /* compute the relative intensity */
      shd = (int)(shdcnt *val); /* and from it the color shade */
      if (shd    >= shdcnt) shd = shdcnt-1;
      if (shd    <  0)      shd = 0;
      if (colcnt <= 0)      col = 0;
      SetPixel(hdc, x, y, shd_pixel(shades, col, shd));
    }                           /* draw a pixel in the */
  }                             /* computed color shade */

  /* --- get color and pen for markers --- */
  if ((colcnt > 0) && (shd_shdbase(shades) == SHD_BLACK))
       { mrk = RGB(255,255,255); SelectObject(hdc, p_white); }
  else { mrk = RGB(  0,  0,  0); SelectObject(hdc, p_black); }

  /* --- draw markers --- */
  if (mrkflgs & (MRK_CENTER|MRK_ELL1S|MRK_ELL2S|MRK_ELL3S)) {
    for (i = clscnt; --i >= 0; ) {   /* traverse the classes */
      if (colcnt <= 0)          /* on grey use the class color */
        mrk = shd_pixel(datcols, i, 0);
      rx = bvn_ex(bvnorm +i);   /* get the expected values */
      ry = bvn_ey(bvnorm +i);   /* (coordinates of the center) */
      if (mrkflgs & MRK_CENTER){/* if to mark the centers */
        x = (int)((rx -xoff) /xscl +0.5);
        y = (int)((yoff -ry) /yscl +0.5);
        SetPixel(hdc, x-1, y-1, mrk);
        SetPixel(hdc, x,   y-1, mrk);
        SetPixel(hdc, x+1, y-1, mrk);
        SetPixel(hdc, x-1, y,   mrk);
        SetPixel(hdc, x+1, y,   mrk);
        SetPixel(hdc, x-1, y+1, mrk);
        SetPixel(hdc, x,   y+1, mrk);
        SetPixel(hdc, x+1, y+1, mrk);
      }                         /* draw a 3x3 pixel rectangle */
      dx = bvn_dx(bvnorm +i);   /* get the standard deviations */
      dy = bvn_dy(bvnorm +i);   /* (extensions of the ellipse) */
      for (k = 1; k <= 3; k++){ /* traverse the ellipses */
        if (!(mrkflgs & (MRK_ELL1S << (k-1))))
          continue;             /* skip unselected ellipses */
        cmin = (int)((rx -k*dx -xoff) /xscl +0.5);
        cmax = (int)((rx +k*dx -xoff) /xscl +0.5);
        for (x = cmin; x <= cmax; x++) {
          t0 = xoff +x *xscl;
          bvn_ellx2y(bvnorm +i, k, t0, &t1, &t2);
          y = (int)((yoff -t1) /yscl +0.5);
          SetPixel(hdc, x, y, mrk);
          y = (int)((yoff -t2) /yscl +0.5);
          SetPixel(hdc, x, y, mrk);
        }
        cmin = (int)((yoff -ry -k*dy) /yscl +0.5);
        cmax = (int)((yoff -ry +k*dy) /yscl +0.5);
        for (y = cmin; y <= cmax; y++) {
          t0 = yoff -y *yscl;
          bvn_elly2x(bvnorm +i, k, t0, &t1, &t2);
          x = (int)((t1 -xoff) /xscl +0.5);
          SetPixel(hdc, x, y, mrk);
          x = (int)((t2 -xoff) /xscl +0.5);
          SetPixel(hdc, x, y, mrk);
        }                       /* for each k sigma ellipse: */
      }                         /* draw the ellipse twice, */
    }                           /* once traversing the x-coordinates */
  }  /* if (mrkflgs .. */       /* and once the y-coordinates */

  /* --- draw data --- */
  if (!table                    /* if no data table has been loaded */
  ||  !(mrkflgs & MRK_DATA))    /* and the marker flag is not set */
    return;                     /* abort the function */
  if      (nbc) trgid = nbc_trgid(nbc);
  else if (fbc) trgid = fbc_trgid(fbc);
  else          trgid = -1;     /* get the class attribute id. */
  xid   = map_fwd[selatt.h_att];/* get other attribute ids. */
  yid   = map_fwd[selatt.v_att];/* and attribute types */
  xtype = att_type(as_att(attset, xid));
  ytype = att_type(as_att(attset, yid));
  s = mrksize +2; o = s >> 1;   /* compute the marker size parameters */
  for (k = tab_tplcnt(table); --k >= 0; ) {
    tpl = tab_tpl(table, k);    /* traverse the tuples */
    col = (trgid >= 0) ? tpl_colval(tpl, trgid)->i : NV_NOM;
    if (xtype == AT_INT) {      /* if x-coordinate is integer valued */
      i = tpl_colval(tpl, xid)->i;
      if (isnull(i)) continue;
      rx = (double)i; }         /* get and check the attribute value */
    else {                      /* if x-coordinate is real valued */
      f = tpl_colval(tpl, xid)->f;
      if (isnan(f))  continue;
      rx = (double)f;           /* get and check the attribute value */
    }
    if (ytype == AT_INT) {      /* if y-coordinate is integer valued */
      i = tpl_colval(tpl, yid)->i;
      if (isnull(i)) continue;
      ry = (double)i; }         /* get and check the attribute value */
    else {                      /* if y-coordinate is real valued */
      f = tpl_colval(tpl, yid)->f;
      if (isnan(f))  continue;
      ry = (double)f;           /* get and check the attribute value */
    }                           /* then compute the screen coords. */
    x = (int)((rx -xoff) /xscl +0.5) -o;
    y = (int)((yoff -ry) /yscl +0.5) -o;
    brush = ((colcnt > 0) && !(mrkflgs & MRK_COLOR))
         || ((col < 0) && (col >= shd_colcnt(datcols)))
          ? br_grey : brushes[col];
    SelectObject(hdc, brush);   /* get and set the brush for drawing */
    Ellipse(hdc, x, y, x+s, y+s);
  }                             /* draw a circle at the data point */
}  /* redraw() */

/*----------------------------------------------------------------------
  File Selector Function
----------------------------------------------------------------------*/

static const char* fselect (HWND hwnd, char *title)
{                               /* --- select a file name */
  static char fname[_MAX_PATH+1] = "";  /* file name */
  static char path [_MAX_PATH+1];       /* and path */
  static char filter[] =        /* file selection filter */
    "all files (*.*)\0*.*\0";   /* (filter for all files) */
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select File...",           /* lpstrTitle (with default init.) */
    OFN_PATHMUSTEXIST, 0, 0,    /* Flags, nFileOffset, nFileExtension */
    NULL, 0L,                   /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */

  ofn.hwndOwner  = hwnd;        /* set parent window handle */
  ofn.lpstrTitle = title;       /* and dialog box title */
  ofn.Flags      = OFN_FILEMUSTEXIST;
  if (GetOpenFileName(&ofn) == 0)
    return NULL;                /* open file selector box */
  return fname;                 /* return name of selected file */
}  /* fselect() */

/*----------------------------------------------------------------------
  Dialog Box Procedures
----------------------------------------------------------------------*/

BOOL CALLBACK dp_format (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- characters dialog box proc. */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      chk_set(hwnd, DI_NAMES,   datafmt.names);
      edt_set(hwnd, DI_RECORD,  datafmt.recseps);
      edt_set(hwnd, DI_FIELD,   datafmt.fldseps);
      edt_set(hwnd, DI_BLANK,   datafmt.blanks);
      edt_set(hwnd, DI_NULL,    datafmt.nullchs);
      edt_set(hwnd, DI_COMMENT, datafmt.comment);
      return 1;                 /* set defaults from parameter block */

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          datafmt.names = chk_get(hwnd, DI_NAMES) ? 1 : 0;
          edt_get(hwnd, DI_RECORD,  datafmt.recseps, 80);
          edt_get(hwnd, DI_FIELD,   datafmt.fldseps, 80);
          edt_get(hwnd, DI_BLANK,   datafmt.blanks,  80);
          edt_get(hwnd, DI_NULL,    datafmt.nullchs, 80);
          edt_get(hwnd, DI_COMMENT, datafmt.comment, 80);
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_format() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_colors (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- colors dialog box procedure */
  int clscnt;                   /* number of classes/clusters */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      LoadString(hinst, CS_DT_PROB,  buf, sizeof(buf));
      cmb_add(hwnd, DI_DIST, buf);
      LoadString(hinst, CS_DT_FUZZY, buf, sizeof(buf));
      cmb_add(hwnd, DI_DIST, buf);
      LoadString(hinst, CS_DT_POSS,  buf, sizeof(buf));
      cmb_add(hwnd, DI_DIST, buf);      /* set distribution type */
      cmb_set(hwnd, DI_DIST, colinfo.dist);
      LoadString(hinst, CS_SM_COLOR, buf, sizeof(buf));
      cmb_add(hwnd, DI_SHDMODE, buf);
      LoadString(hinst, CS_SM_GREY,  buf, sizeof(buf));
      cmb_add(hwnd, DI_SHDMODE, buf);   /* set shading mode */
      cmb_set(hwnd, DI_SHDMODE, (colinfo.colcnt > 0) ? 0 : 1);
      sprintf(buf, "%g", colinfo.coloff);
      edt_set(hwnd, DI_COLOFF, buf);    /* set color offset */
      LoadString(hinst, CS_SB_BLACK, buf, sizeof(buf));
      cmb_add(hwnd, DI_SHDBASE, buf);
      LoadString(hinst, CS_SB_WHITE, buf, sizeof(buf));
      cmb_add(hwnd, DI_SHDBASE, buf);   /* set shade base */
      cmb_set(hwnd, DI_SHDBASE, (colinfo.shdbase == SHD_BLACK) ? 0 : 1);
      int_set(hwnd, DI_SHDCNT,  colinfo.shdcnt);
      sprintf(buf, "%g", colinfo.ndwgt);/* set number of shades */
      edt_set(hwnd, DI_NDWGT, buf);     /* set weight of distribution */
      return 1;

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          colinfo.dist = (int)cmb_get(hwnd, DI_DIST);
          if      (nbc) clscnt = nbc_clscnt(nbc);
          else if (fbc) clscnt = fbc_clscnt(fbc);
          else          clscnt = -1;    /* get number of classes */
          colinfo.colcnt  = (cmb_get(hwnd, DI_SHDMODE) == 0)
                          ? clscnt : 0; /* get shading mode */
          edt_get(hwnd, DI_COLOFF, buf, 80);
          colinfo.coloff  = atof(buf);  /* get color offset (angle) */
          colinfo.shdbase = (cmb_get(hwnd, DI_SHDBASE) == 0)
                          ? SHD_BLACK : SHD_WHITE;
                                        /* get shade base */
          colinfo.shdcnt  = int_get(hwnd, DI_SHDCNT);
          if (colinfo.shdcnt < 1)          colinfo.shdcnt = 1;
          if (colinfo.shdcnt > SHD_MAXCNT) colinfo.shdcnt = SHD_MAXCNT;
                                        /* get number of shades */
          edt_get(hwnd, DI_NDWGT, buf, 80);
          colinfo.ndwgt   = atof(buf);  /* get weight of Gaussian */
          if (colinfo.ndwgt < 0) colinfo.ndwgt  = 0;
          if (colinfo.ndwgt > 1) colinfo.ndwgt  = 1;
          colinfo.cowgt   = 1 -colinfo.ndwgt;
          if (col_create() != 0)/* create new color shades */
            error(E_COLOR);     /* and marker colors */
          EndDialog(hwnd, 0);   /* and terminate the dialog */
          return 1;
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_colors() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_marks (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- markers dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      chk_set(hwnd, DI_DATA,   (mrkflgs & MRK_DATA)   ? 1 : 0);
      chk_set(hwnd, DI_COLOR,  (mrkflgs & MRK_COLOR)  ? 1 : 0);
      chk_set(hwnd, DI_CENTER, (mrkflgs & MRK_CENTER) ? 1 : 0);
      chk_set(hwnd, DI_ELL1S,  (mrkflgs & MRK_ELL1S)  ? 1 : 0);
      chk_set(hwnd, DI_ELL2S,  (mrkflgs & MRK_ELL2S)  ? 1 : 0);
      chk_set(hwnd, DI_ELL3S,  (mrkflgs & MRK_ELL3S)  ? 1 : 0);
      sprintf(buf, "%d", mrksize); edt_set(hwnd, DI_SIZE, buf);
      return 1;                 /* set defaults from marker flags */

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          if (chk_get(hwnd, DI_DATA))   mrkflgs |=  MRK_DATA;
          else                          mrkflgs &= ~MRK_DATA;
          if (chk_get(hwnd, DI_COLOR))  mrkflgs |=  MRK_COLOR;
          else                          mrkflgs &= ~MRK_COLOR;
          if (chk_get(hwnd, DI_CENTER)) mrkflgs |=  MRK_CENTER;
          else                          mrkflgs &= ~MRK_CENTER;
          if (chk_get(hwnd, DI_ELL1S))  mrkflgs |=  MRK_ELL1S;
          else                          mrkflgs &= ~MRK_ELL1S;
          if (chk_get(hwnd, DI_ELL2S))  mrkflgs |=  MRK_ELL2S;
          else                          mrkflgs &= ~MRK_ELL2S;
          if (chk_get(hwnd, DI_ELL3S))  mrkflgs |=  MRK_ELL3S;
          else                          mrkflgs &= ~MRK_ELL3S;
          edt_get(hwnd, DI_SIZE, buf, 80);
          mrksize = atoi(buf); if (mrksize <= 0) mrksize = 1;
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_marks() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_atts (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- attributes dialog box proc. */
  int i;                        /* loop variable, buffer */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      if (names) {              /* if there are attribute names */
        for (i = 0; i < attcnt; i++) cmb_add(hwnd, DI_H_ATT, names[i]);
        cmb_set(hwnd, DI_H_ATT, selatt.h_att);
      }                         /* add them to the list box */
      sprintf(buf, "%g", selatt.h_rng.min);  /* set minimal value */
      edt_set(hwnd, DI_H_MIN, buf);    /* for horizontal direction */
      sprintf(buf, "%g", selatt.h_rng.max);  /* set maximal value */
      edt_set(hwnd, DI_H_MAX, buf);    /* for horizontal direction */
      if (names) {              /* if there are attribute names */
        for (i = 0; i < attcnt; i++) cmb_add(hwnd, DI_V_ATT, names[i]);
        cmb_set(hwnd, DI_V_ATT, selatt.v_att);
      }                         /* add them to the list box */
      sprintf(buf, "%g", selatt.v_rng.min);  /* set minimal value */
      edt_set(hwnd, DI_V_MIN, buf);      /* for vertical direction */
      sprintf(buf, "%g", selatt.v_rng.max);  /* set maximal value */
      edt_set(hwnd, DI_V_MAX, buf);      /* for vertical direction */
      return 1;                 /* set defaults from parameter block */

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (LOWORD(wp)) {     /* evaluate command code */

        case DI_H_ATT:          /* if attribute for horizontal dir. */
          if ((HIWORD(wp) != CBN_SELCHANGE) || !ranges)
            break;              /* ignore anything but sel. change */
          i = (int)cmb_get(hwnd, DI_H_ATT);   /* get new selection */
          sprintf(buf, "%g", ranges[i].min);  /* set minimal value */
          edt_set(hwnd, DI_H_MIN, buf);       /* for horizontal dir. */
          sprintf(buf, "%g", ranges[i].max);  /* set maximal value */
          edt_set(hwnd, DI_H_MAX, buf);       /* for horizontal dir. */
          break;

        case DI_V_ATT:          /* if attribute for vertical dir. */
          if ((HIWORD(wp) != CBN_SELCHANGE) || !ranges)
            break;              /* ignore anything but sel. change */
          i = (int)cmb_get(hwnd, DI_V_ATT);   /* get new selection */
          sprintf(buf, "%g", ranges[i].min);  /* set minimal value */
          edt_set(hwnd, DI_V_MIN, buf);       /* for horizontal dir. */
          sprintf(buf, "%g", ranges[i].max);  /* set maximal value */
          edt_set(hwnd, DI_V_MAX, buf);       /* for horizontal dir. */
          break;

        case IDOK:              /* if 'Ok' button pressed */
          selatt.h_att = (int)cmb_get(hwnd, DI_H_ATT);
          edt_get(hwnd, DI_H_MIN, buf, 80);   /* get horz. min. */
          ranges[selatt.h_att].min = selatt.h_rng.min = atof(buf);
          edt_get(hwnd, DI_H_MAX, buf, 80);   /* get horz. max. */
          ranges[selatt.h_att].max = selatt.h_rng.max = atof(buf);
          selatt.v_att = (int)cmb_get(hwnd, DI_V_ATT);
          edt_get(hwnd, DI_V_MIN, buf, 80);   /* get vert. min. */
          ranges[selatt.v_att].min = selatt.v_rng.min = atof(buf);
          edt_get(hwnd, DI_V_MAX, buf, 80);   /* get vert. max. */
          ranges[selatt.v_att].max = selatt.v_rng.max = atof(buf);
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */

        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_atts() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_about (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'about' dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_COMMAND:            /* command (button pressed etc.) */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* dialog box closed */
      EndDialog(hwnd, 0); return 1;
  }                             /* terminate dialog */
  return 0;                     /* everything else not handled */
}  /* dp_about() */

/*----------------------------------------------------------------------
  Window Procedures
----------------------------------------------------------------------*/

LRESULT CALLBACK wp_main (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- main window procedure */
  static PAINTSTRUCT ps;        /* for BeginPaint call */
  static int  szmv = 0;         /* size/move flag */
  const  char *fname;           /* name of selected file */
  HDC    hdc;                   /* handle of device context */
  int    r;                     /* result of dialogs, load functions */

  switch (msg) {                /* evaluate message */
    case WM_GETMINMAXINFO:      /* window is about to be resized */
      ((LPMINMAXINFO)lp)->ptMinTrackSize.x = MINXEXT;
      ((LPMINMAXINFO)lp)->ptMinTrackSize.y = MINYEXT;
      break;                    /* set a minimal width and height */

    case WM_ENTERSIZEMOVE:      /* if window is resized or moved, */
      szmv = 1; break;          /* inhibit redrawing with a flag */
    case WM_SIZING:             /* if window is actually resized, */
      szmv = 2; break;          /* note this in the flag */
    case WM_EXITSIZEMOVE:       /* if resizing/moving is finished */
      if (szmv > 1) InvalidateRect(hwnd, NULL, 0);
      szmv = 0; break;          /* redraw the window contents */

    case WM_PAINT:              /* --- if to paint window */
      if (szmv) break;          /* during resize do not redraw */
      hdc = BeginPaint(hwnd, &ps);
      if (!hdc) return 0;       /* get device context for drawing */
      redraw(hdc);              /* redraw the window contents */
      EndPaint(hwnd, &ps);      /* release the device context */
      break;

    case WM_COMMAND:            /* --- if menu item selected */
      switch (wp) {             /* evaluate item */
        /* --- Menu File --- */
        case MI_CLASS:          /* File > Load Classifier... */
          LoadString(hinst, CS_LD_CLASS, buf, sizeof(buf));
          fname = fselect(hwnd, buf);
          if (!fname) return -1;/* let the user select a file */
          r = bc_load(fname);   /* load a classifier/set of clusters */
          if (r != 0) { error(r, fname); return -1; }
          if (col_create() != 0) {      /* get new color shades */
            error(E_COLOR); return -1; }   /* and marker colors */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* redraw the window contents */

        case MI_DATA:           /* File > Load Data... */
          if (!attset) { error(E_DATA); break; }
          LoadString(hinst, CS_LD_DATA, buf, sizeof(buf));
          fname = fselect(hwnd, buf);
          if (!fname) return -1;/* let the user select a file */
          r = bc_data(fname);   /* load a data table */
          if (r) { error(r, (r > 0) ? bc_msg : fname); return -1; }
          mrkflgs |= MRK_DATA;  /* enforce marks at the data points */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* redraw the window contents */

        case MI_FORMAT:         /* File > Data Format... */
          DialogBox(hinst, MKIRSC(DB_FORMAT), hwnd,(DLGPROC)dp_format);
          return 0;             /* open the dialog box */

        case MI_QUIT:           /* File > Quit */
          DestroyWindow(hwnd);  /* destroy main window */
          return 0;             /* and abort the program */

        /* --- Menu Actions --- */
        case MI_ATTS:           /* Menu Item 'Actions > Attributes...'*/
          if (DialogBox(hinst, MKIRSC(DB_ATTS), hwnd,
             (DLGPROC)dp_atts) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* on ok redraw the window contents */

        case MI_COLORS:         /* Menu Item 'Actions > Colors...' */
          if (DialogBox(hinst, MKIRSC(DB_COLORS), hwnd,
                       (DLGPROC)dp_colors) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* on ok redraw the window contents */

        case MI_MARKS:          /* Menu Item 'Actions > Markers...' */
          if (DialogBox(hinst, MKIRSC(DB_MARKS), hwnd,
             (DLGPROC)dp_marks) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* on ok redraw the window contents */

        case MI_REDRAW:         /* Menu Item 'Actions > Redraw' */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        /* --- Menu Help --- */
        case MI_ABOUT:          /* Menu Item 'Help > About' */
          DialogBox(hinst, MKIRSC(DB_ABOUT), hwnd, (DLGPROC)dp_about);
          return 0;             /* show 'about' dialog box */
      } break;

    case WM_CLOSE:              /* window closed */
    case WM_DESTROY:            /* or destroyed */
      PostQuitMessage(0);       /* terminate program */
      return 0;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}  /* wp_main() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int WINAPI WinMain (HINSTANCE hcurr, HINSTANCE hprev,
                    LPSTR cmdln, int cmd)
{                               /* --- main function */
  WNDCLASS wc;                  /* window class */
  RECT     rc;                  /* window rectangle */
  MSG      msg;                 /* message */
  char     *arg;                /* command line argument */
  int      r;                   /* results of calls to bc_... */
  char     *s;                  /* to traverse the command line */
  int      c;                   /* buffer for a character */

  /* --- create window class --- */
  hinst = hcurr;                /* note program instance handle */
  if (!hprev) {                 /* if first instance of this program */
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = wp_main;
    wc.cbClsExtra    = wc.cbWndExtra = 0;
    wc.hInstance     = hinst;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName  = MKIRSC(MN_MAIN);
    wc.lpszClassName = "WBCViewClass";
    if (!RegisterClass(&wc))    /* register class of */
      return FALSE;             /* main window */
  }

  /* --- get pens and brushes --- */
  p_white = (HPEN)GetStockObject(WHITE_PEN);
  p_black = (HPEN)GetStockObject(BLACK_PEN);
  br_grey = CreateSolidBrush(RGB(160, 160, 160));

  /* --- create main window --- */
  rc.left = 30; rc.right  = 30 +INITXEXT -1;
  rc.top  = 30; rc.bottom = 30 +INITYEXT -1;
  AdjustWindowRect(&rc, WINSTYLE, 1);
  hmain = CreateWindow("WBCViewClass", PRGNAME, WINSTYLE,
                       30, 30, rc.right-rc.left, rc.bottom-rc.top,
                       NULL, NULL, hinst, NULL);
  if (hmain == 0) return FALSE; /* create main window */

  /* --- load initial classifier and data --- */
  while (cmdln && *cmdln) {     /* if a command line is given */
    s = cmdln;                  /* traverse the comand line */
    while (*s == ' ') s++;      /* find start of file name */
    if (*s == '\"') { arg = s+1; c = '\"'; }
    else            { arg = s;   c = ' ';  }
    while ((*s != c) && (*s != 0)) s++;
    c = *s; *s++ = 0;           /* find end of file name */
    if (*arg) { r = bc_load(arg);     /* load classifier */
      if (r != 0) { error(r, arg); break; } }
    if (col_create()) break;    /* create shades and colors */
    if (!c) break;              /* check for another argument */
    while (*s == ' ') s++;      /* find start of file name */
    if (*s == '\"') { arg = s+1; c = '\"'; }
    else            { arg = s;   c = ' ';  }
    while ((*s != c) && (*s != 0)) s++;
    *s = 0;                     /* find end of file name */
    if (*arg) { r = bc_data(arg);     /* load data table */
      if (r) { error(r, (r > 0) ? bc_msg : arg); }
    }                           /* (load specified files) */
    break;                      /* terminate dummy loop */
  }  /* while (cmdln && *cmdln) */

  /* --- show main window --- */
  ShowWindow(hmain, cmd);       /* open and */
  UpdateWindow(hmain);          /* update main window */

  /* --- dispatch messages --- */
  while (GetMessage(&msg, NULL, 0, 0)) {
    TranslateMessage(&msg);     /* get messages, translate */
    DispatchMessage (&msg);     /* and dispatch them */
  }
  bc_clean();                   /* clean up variables */
  col_delete();                 /* and delete colors/shades */
  return (int)msg.wParam;
}  /* WinMain() */
