/*----------------------------------------------------------------------
  File    : matrix3.c
  Contents: general vector and matrix management
            eigenvalue and eigenvector functions
  Author  : Christian Borgelt
  History : 1999.05.18 file created
            2004.02.10 function mat_jacobi() completed
            2004.02.11 function mat_3dred() completed
            2004.02.12 function mat_3dqli() completed
            2007.11.22 function mat_evsort() added (eigenvalue sorting)
            2013.08.09 adapted to higher compiler warning level
            2013.08.13 adapted to preprocessor definition of DIMID
----------------------------------------------------------------------*/
#include <stdio.h>
#include <limits.h>
#include <float.h>
#include <assert.h>
#include "matrix.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/
#if !defined _MSC_VER && !defined _hypot

static double _hypot (double a, double b)
{                               /* --- length of hypotenuse */
  a = fabs(a); b = fabs(b);     /* compute the absolute values */
  if (a > b) { b /= a; return a *sqrt(1 +b*b); }
  else       { a /= b; return b *sqrt(1 +a*a); }
}  /* _hypot() */               /* compute the hypotenuse length */
                                /* avoiding underflow or overflow */
#endif
/*----------------------------------------------------------------------
  Jacobi Transformations of a Symmetric Matrix

  Reference:
  W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P. Flannery.
  Numerical Recipes in C --- The Art of Scientific Computing
  (2nd edition), pp. 463--468.
  Cambridge University Press, Cambridge, United Kingdom 1992
----------------------------------------------------------------------*/

int mat_jacobi (double *evs, MATRIX *res, MATRIX *mat, int max)
{                               /* --- Jacobi transform. sym. matrix */
  DIMID  row, col, i;           /* loop variables */
  int    rotcnt = 0;            /* number of Jacobi rotations */
  double *r, *r1, *r2;          /* to traverse/access matrix rows */
  double *chg;                  /* changes of diagonal elements */
  double sum;                   /* sum of off-diagonal elements */
  double tcf;                   /* coefficient for threshold */
  double c, d, s, t, e;         /* temporary buffers */

  assert(mat && (res || evs)    /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (!res || ((res->rowcnt == mat->rowcnt)
              &&  (res->colcnt == mat->colcnt))));
  if (res) {                    /* if eigenvectors are requested */
    for (row = mat->rowcnt; --row >= 0; ) {
      for (r = res->els[row] +(col = mat->colcnt); --col >= 0; )
        *--r = 0;               /* initialize the eigenvectors */
      r[row] = 1;               /* (matrix columns) to unit vectors */
    }                           /* (unit transformation matrix) */
    if (!evs) evs = res->vec;   /* use the associated vector */
  }                             /* for the eigenvalues if necessary */
  chg = res->buf;               /* get the internal buffer */
  for (i = mat->rowcnt; --i >= 0; ) {
    evs[i] = mat->els[i][i];    /* copy the diagonal elements */
    chg[i] = 0;                 /* and initialize the changes */
  }                             /* (separate for accuracy reasons) */
  tcf = 0.2 /((double)mat->rowcnt *(double)mat->colcnt);
  while (--max >= 0) {          /* do Jacobi sweeps over the matrix */
    for (sum = 0, row = mat->rowcnt; --row > 0; )
      for (r = mat->els[row] +(col = row); --col >= 0; )
        sum += fabs(*--r);      /* sum off-diagonal elements */
    if (!(sum > 0)) break;      /* and check for termination */
    sum = (i < 4) ? sum *tcf : 0;
    for (row = mat->rowcnt; --row > 0; ) {
      r = mat->els[row];        /* get the matrix row */
      for (col = row; --col >= 0; ) {
        s = evs[row] +chg[row]; /* traverse the lower triangle */
        t = evs[col] +chg[col]; /* (row/column combinations) */
        e = 100 *r[col];        /* after four matrix traversals */
        if (i > 4) {            /* skip rotation for small elements */
               c =  fabs(s) +e;  d =  fabs(t) +e;
          if ((c == fabs(s)) && (d == fabs(t))) {
            r[col] = 0; continue; }
        }
        if (fabs(r[col]) <= sum)/* in the first four sweeps, skip */
          continue;             /* rotation based on other criterion */
        d = s -t;               /* difference of diagonal elements */
        t = fabs(d) +e;         /* check size of e relative to d */
        if (t == fabs(d))       /* if an overflow is to be expected, */
          t = r[col] /d;        /* t = 1/(2\theta) = 1/(2\cot(2\phi)) */
        else {                  /* if no overflow is to be expected, */
          s = 0.5*d /r[col];    /* use smaller solution of equation */
          t = ((s < 0) ? -1 : 1) /(fabs(s) +sqrt(1+s*s));
        }                       /* t^2 +2t\theta -1 = 0 (\theta = s) */
        c = 1 /sqrt(1 +t*t);    /* compute the cosine and sine */
        s = t*c;                /* of the rotation angle \phi and the */
        c = s /(1+c);           /* tangens of half the rotation angle */
        chg[col] -= t *= r[col];
        chg[row] += t;          /* update changes of diagonal elems. */
        r[col]    = 0;          /* off-diagonal element is now zero */
        r1 = mat->els[col];     /* update the other row/column elems. */
        r2 = mat->els[row];     /* according to the following schemes */
        for (i = 0; i < col; i++) {  /*       col row   */
          d       = r1[i];           /*        |   |    */
          t       = r2[i];           /* col ###+---+--- */
          r1[i]   = d -s *(t +d*c);  /*        |   |    */
          r2[i]   = t +s *(d -t*c);  /* row ###+---+--- */
        }                            /*        |   |    */
        for (i = col+1; i < row; i++) {
          r1      = mat->els[i];     /*       col row   */
          d       = r1[col];         /*        |   |    */
          t       = r2[i];           /* col ---+---+--- */
          r1[col] = d -s *(t +d*c);  /*        #   |    */
          r2[i]   = t +s *(d -t*c);  /* row ---+###+--- */
        }                            /*        |   |    */
        for (i = row+1; i < mat->rowcnt; i++) {
          r1      = mat->els[i];     /*       col row   */
          d       = r1[col];         /*        |   |    */
          t       = r1[row];         /* col ---+---+--- */
          r1[col] = d -s *(t +d*c);  /*        |   |    */
          r1[row] = t +s *(d -t*c);  /* row ---+---+--- */
        }                            /*        #   #    */
        rotcnt++;               /* count the Jacobi rotation */
        if (!res)               /* if no eigenvectors are requested, */
          continue;             /* skip rotation of the result matrix */
        for (i = mat->rowcnt; --i >= 0; ) {
          r1      = res->els[i];     /* update the eigenvectors */
          d       = r1[col];         /* (columns of result matrix) */
          t       = r1[row];         /*       col row   */
          r1[col] = d -s *(t +d*c);  /*        #   #    */
          r1[row] = t +s *(d -t*c);  /* col ---#---#--- */
        }                            /*        #   #    */
      }                              /* row ---#---#--- */
    }                                /*        #   #    */
    for (i = mat->rowcnt; --i >= 0; ) {
      evs[i] += chg[i];         /* update the eigenvalues */
      chg[i]  = 0;              /* and reinitialize the changes */
    }                           /* (aggregate the changes first */
  }                             /* to achieve a higher accuracy) */
  return (max < 0) ? -2 : rotcnt;
}  /* mat_jacobi() */           /* return the number of rotations */

/*----------------------------------------------------------------------
The above function leaves the upper triangle of the matrix (including
the diagonal) as well as the associated vector intact and works only
on the lower triangle. That is, the original symmetric matrix can
always be reconstructed by a call to mat_tr2sym(mat, mat, MAT_UPPER).
This removes the necessity to copy the matrix or the associated vector
before the call. It may, however, be necessary to copy the upper
triangle to the lower before the call, because functions like, for
instance, mat_addmp(), mat_covar() etc. only set the upper triangle.
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  Householder Reduction of a Symmetric Matrix to Tridiagonal Form

  Reference:
  W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P. Flannery.
  Numerical Recipes in C --- The Art of Scientific Computing
  (2nd edition), pp. 469--475.
  Cambridge University Press, Cambridge, United Kingdom 1992
----------------------------------------------------------------------*/

void mat_3dred (double *dia, double *sub, MATRIX *otm,
                MATRIX *mat, int loc)
{                               /* --- Householder red. to 3-diagonal */
  DIMID  row, col, i;           /* loop variables */
  double *r, *p;                /* to traverse/access matrix rows */
  double *b;                    /* to access the matrix buffer */
  double f, g, h, s;            /* temporary buffers */
  int    t = 1;                 /* flag for transformation matrix */

  assert(mat && dia && sub      /* check the function arguments */
  &&    (mat->rowcnt == mat->colcnt)
  &&    (!otm || ((otm->rowcnt == mat->rowcnt)
              &&  (otm->colcnt == mat->colcnt))));
  if (!otm) {                   /* if no transformation matrix */
    otm = mat; t = 0; }         /* is requested, work on input */
  if      (loc == MAT_UPPER) {  /* if only upper triangle is valid */
    for (row = otm->rowcnt; --row >= 0; ) {
      col = (otm != mat) ? row+1 : row;
      for (r = otm->els[row] +col; --col >= 0; )
        *--r = mat->els[col][row];
    } }                         /* copy upper to lower triangle */
  else if (otm != mat) {        /* if result differs from source */
    for (row = otm->rowcnt; --row >= 0; ) {
      r = mat->els[row] +(col = row +1);
      p = otm->els[row] + col;
      while (--col >= 0) *--p = *--r;
    }                           /* copy the lower triangle of the */
  }                             /* source to the destination */
  b = otm->buf;                 /* get the internal buffer */
  for (row = otm->rowcnt; --row > 0; ) {
    r = otm->els[row]; s = 0;   /* traverse the matrix rows */
    if (row > 1) {              /* if this is not the first row */
      for (col = row; --col >= 0; )
        s += fabs(r[col]);      /* sum elements before diagonal */
    }                           /* to obtain the scaling factor */
    if (s <= 0) {               /* if row is empty, skip transform. */
      sub[row-1] = r[row-1]; dia[row] = 0; continue; }
    for (h = f = 0, col = 0; col < row; col++) {
      f  = r[col] /= s;         /* sum the squares of the elements */
      h += f*f;                 /* to obtain the squared length */
    }                           /* of the Householder vector u */
    g = (f > 0) ? -sqrt(h) : sqrt(h);
    dia[row]   = h -= f*g;      /* compute and store h = 1/2 |u|^2 */
    sub[row-1] = s*g;           /* store the subdiagonal element */
    r[row-1]   = f-g;           /* store vector u in the matrix row */
    for (f = 0, col = 0; col < row; col++) {
      p = otm->els[col];        /* traverse the matrix columns */
      if (t)                    /* if transform. matrix requested, */
        p[row] = r[col] /h;     /* store u/h in the matrix column */
      for (g = 0, i = 0; i <= col; i++)
        g += p[i] *r[i];        /* g = col-th element of A*u */
      for (i = col+1;    i < row; i++)
        g += otm->els[i][col] *r[i];
      b[col] = s = g/h;         /* store col-th element of */
      f += s *r[col];           /* p = (A*u)/h in sub[col] */
    }                           /* and compute f = u^T*p */
    s = f /(h+h);               /* compute s = (u^T*p)/2h */
    for (col = 0; col < row; col++) {
      f = r[col];               /* form the elements of q = p -s*u */
      g = b[col] -= s*f;        /* and store them in the vector sub */
      p = otm->els[col];        /* overwriting the elements of p */
      for (i = 0; i <= col; i++)
        p[i] -= f *b[i] +g *r[i];
    }                           /* compute A' = A -q*u^T -u*q^T */
  }                             /* (i.e., reduce the matrix A) */
  if (!t) {                     /* if no transform. matrix requested */
    for (row = otm->rowcnt; --row >= 0; )
      dia[row] = otm->els[row][row]; }  /* only copy the diagonal */
  else {                        /* if transform. matrix requested */
    dia[0] = 0;                 /* accumulate transform. matrices */
    for (row = 0; row < otm->rowcnt; row++) {
      r = otm->els[row];        /* traverse the matrix rows */
      if (dia[row] != 0) {      /* skip computation for zero diagonal */
        for (col = 0; col < row; col++) {
          for (g = 0, i = 0; i < row; i++)
            g += r[i] *otm->els[i][col];
          for (i = 0; i < row; i++) {
            p = otm->els[i]; p[col] -= g *p[row]; }
        }                       /* use stored vectors u (row) and */
      }                         /* and u/H (column) to form P*Q */
      dia[row] = r[row];        /* copy the diagonal element */
      for (col = 0; col < row; col++)
        otm->els[col][row] = r[col] = 0;
      r[row] = 1;               /* reset row and column to identity */
    }                           /* matrix for the next iteration */
  }
}  /* mat_3dred() */

/*----------------------------------------------------------------------
The above function works only on the lower triangle of the input matrix,
except when result and input matrix are identical or no result matrix is
given. On completion, res contains the orthogonal transformation matrix
that accomplishes the reduction to tridiagonal form. diag contains the
diagonal, sub the subdiagonal of the tridiagonal matrix.
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  QL Algorithm with Implicit Shifts for a Tridiagonal Matrix

  Reference:
  W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P. Flannery.
  Numerical Recipes in C --- The Art of Scientific Computing
  (2nd edition), pp. 475--481.
  Cambridge University Press, Cambridge, United Kingdom 1992
----------------------------------------------------------------------*/

int mat_3dqli (double *evs, MATRIX *mat,
               const double *dia, double *sub, DIMID n, DIMID max)
{                               /* --- QL alg. with implicit shifts */
  DIMID  row, i, k, iter;       /* loop variables */
  DIMID  spr;                   /* index of row to split matrix at */
  double *r;                    /* to traverse/access a matrix row */
  double c, s;                  /* sine and cosine of rotation angle */
  double tc, ts;                /* sine/cosine * subdiagonal element */
  double g, h, p;               /* temporary buffers */

  assert(dia && sub             /* check the function arguments */
  &&    ( mat || evs) && (mat || (n > 0))
  &&    (!mat || (n <= 0) || (n == mat->rowcnt))
  &&    (!mat || (mat->rowcnt == mat->colcnt)));
  if (mat) {                    /* if eigenvectors are requested */
    n = mat->rowcnt;            /* get the matrix dimension */
    r = mat->buf;               /* and the internal buffer */
    for (i = n; --i >= 0; )     /* copy the subdiagonal to the */
      r[i] = sub[i];            /* internal buffer and switch to it */
    sub = r;                    /* to preserve the input */
    if (!evs) evs = mat->vec;   /* use the associated vector */
  }                             /* for the eigenvalues if necessary */
  sub[n-1] = 0;                 /* clear the add. subdiagonal element */
  if (evs != dia) {             /* if not to work in place, */
    for (i = n; --i >= 0; )     /* copy the matrix diagonal */
      evs[i] = dia[i];          /* to the eigenvalue vector */
  }                             /* (the diagonal will be modified) */
  if (max <= 0) max = n;        /* check the number of iterations */
  for (row = 0; row < n-1; row++) {
    iter = 0;                   /* initialize the iteration counter */
    while (1) {                 /* process submatrices */
      for (spr = row; spr < n-1; spr++) {
        g = fabs(evs[spr]) +fabs(evs[spr+1]);
        h = fabs(sub[spr]) +g;  /* find a small subdiagonal element */
        if (g == h) break;      /* for splitting the matrix */
      }                         /* into submatrices */
      if (spr == row) break;    /* if submatrix is empty, abort */
      if (++iter > max)         /* if too many iterations, */
        return -1;              /* abort the function with failure */
      g = (evs[row+1]-evs[row]) /* solve 2x2 eigenvalue problem */
        / (2 *sub[row]);        /* (see below for more details) and */
      h = _hypot(g, 1);         /* thus compute the shift value k_s */
      if (g < 0) h = -h;        /* (eigenvalue closer to evs[row]) */
      h = evs[row] -sub[row] /(g+h);
      g = evs[spr] -h;          /* h = k_s, g = dia[spr] -k_s */
      s = c = 1;                /* initialize sine and cosine */
      p = 0;                    /* for the plane rotations */
      for (i = spr; --i >= row; ) {     /* do QL and Givens rotations */
        ts = s *sub[i];         /* compute elems. (i+1,i) and (i+2,i) */
        tc = c *sub[i];         /* resulting from preceding rotation */
        h  = _hypot(ts, g);     /* compute new denominator for s, c */
        if (i+1 < n-1)          /* if within range of subdiagonal */
          sub[i+1] = h;         /* store the rot. result (move down) */
        if (h == 0) {           /* on underflow abort the loop */
          evs[i+1] -= p; break; }        /* (end of submatrix) */
        s = ts/h;               /* compute the sine and cosine */
        c = g /h;               /* of the next rotation angle */
        g = evs[i+1] -p;        /* compute new diagonal element */
                                /* (orig. at dia[spr], but moved up) */
        h = (evs[i] -g) *s +2*c*tc;
        p = s*h;                /* compute change of diagonal element */
        evs[i+1] = g+p;         /* store new dia[i] in dia[i+1] */
                                /* (element at dia[spr] is moved up) */
        g = c*h -tc;            /* compute new element sub[i] */
        if (!mat)               /* if no eigenvectors are requested, */
          continue;             /* skip rotation of eigenvectors */
        for (k = 0; k < n; k++) {
          r = mat->els[k]; ts = r[i+1];
          r[i+1] = s*r[i]+c*ts; /* apply the rotation carried out */
          r[i  ] = c*r[i]-s*ts; /* above on the tridiagonal matrix */
        }                       /* to the given matrix */
      }                         /* (unit matrix or otm of mat_3dred) */
      sub[spr]  = 0;            /* zero the subdiagonal split element */
      if (i >= row) continue;   /* recovery from underflow */
      evs[row] -= p;            /* compute changed element dia[spr] */
      sub[row]  = g;            /* and store changed element sub[spr] */
    }                           /* (which were moved up, while all */
    /* fprintf(stderr, "iter: %d, max: %d\n", iter, max); */
  }                             /* other elements were moved down) */
  return 0;                     /* return 'ok' */
}  /* mat_3dqli() */

/*----------------------------------------------------------------------
The above function preserves the subdiagonal sub if a matrix mat is
given. The matrix mat should be a unit matrix if the eigenvalues and
eigenvectors of a tridiagonal matrix are to be found directly and it
should be the output of the function mat_3dred() if the eigenvalues
of a general symmetric matrix are to be found.

Details about the computation of the shift value k_s:
k_s is computed from the solution of the 2x2 eigenvalue problem
/ a c \   where a and b are the diagonal elements at indices row
\ c b /   and row+1 and c is the corresponding subdiagonal element,
selecting the solution that is closer to the diagonal element a.
The solution is computed via the characteristic polynomial
  (a -\lambda)(b -\lambda) -c^2 = 0
However, to make sure that the obtained eigenvalue is the one closer
to the diagonal element a, the quadratic equation is rewritten as
  1/(2c) x*(b-a+x)         -c/2 = 0  <=>
  1/(2c) x^2 +(b-a)/(2c) x -c/2 = 0
with x = a -\lambda (and thus \lambda = a -x). Then one solves using
  x_1 = 2c*q   and   x_2 = -c/(2q),   where
  q = -1/2 *[z +sgn(z)\sqrt{z^2 +1}] and z = (b-a)/(2c).
Here x_2 has the smaller absolute value and thus we have
  k_s = a -x = a +c/(2q) = a -c/[z +sgn(z)\sqrt{z^2 +1}],
which is the formula used in the function above.

Reference for the quadratic equation solution:
  W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P. Flannery.
  Numerical Recipes in C --- The Art of Scientific Computing
  (2nd edition), pp. 183--184.
  Cambridge University Press, Cambridge, United Kingdom 1992

One traversal of the for-loop "for (i = spr; --i >= row; ) ..."
does the following to a 3x3 submatrix:
/ d1 e1 0  \        / d1  e1' e2' \        / d1  e2' e1' \
| e1 d2 e2 | --1--> | e1' d2' 0   | --2--> | e2' d3' 0   |
\ 0  e2 d3 /        \ e2' 0   d3' /        \ e1' 0   d2' /
Where step 1 is a simple plane rotation to annihilate the subdiagonal
elements (computed as for Jacobi transformation) and step 2 exchanges
rows and columns. Formally the second step can be achieved by
multiplying with the orthogonal transformation matrix
    / 1 0 0 \
P = | 0 0 1 |   to obtain    A' = P^T A P.
    \ 0 1 0 /
Note, however, that this is not carried out explicitely, but implicitly
by shifting up the diagonal element d3' and the subdiagonal element e2',
while all other elements are shifted down. After the two steps shown
above, the submatrix looked at is shifted up and left by one element
and the process repeats until the upper boundary of the section with
non-zero subdiagonal elements is encountered. Then no plane rotation
is necessary to annihilate the off-diagonal elements and an exchange of
rows and columns in the upper left 2x2 submatrix restores tridiagonal
form (this is achieved by the statements at the end of the for-loop).

Note that the exchange of rows and columns is reflected in the rotation
of the eigenvector matrix in the fact that the matrix P is already
applied to the plane rotation matrix (which is why the sines appear
in the main diagonal of the rotation matrix).
----------------------------------------------------------------------*/

void mat_evsort (double *evs, MATRIX *mat, int dir)
{                               /* --- sort eigenvalues and -vectors */
  DIMID  i, s, d, n;            /* loop variables, indices */
  DIMID  h;                     /* increment for index */
  double v;                     /* buffer for eigenvalue  to insert */
  double *r, **vecs;            /* buffer for eigenvector to insert */

  assert(!mat || (mat->colcnt == mat->rowcnt));
  if (!evs) evs = mat->vec;     /* get eigenvalues if necessary */
  if (mat) {                    /* if eigenvectors are given */
    mat_transp(mat, mat);       /* turn eigenvectors into rows */
    vecs = mat->els;            /* get the matrix elements and */
    n    = mat->colcnt; }       /* the number of eigenvalues */
  else {                        /* if no eigenvectors are given */
    vecs = NULL;                /* clear the vector buffer and */
    n    = abs(dir);            /* get the number of values */
  }                             /* from the sorting direction */
  r = NULL;                     /* initialize to avoid warning */
  for (h = n; h > 1; ) {        /* traverse the increments */
    h = (h < 5) ? 1 : ((h *5 -1) /11);
    for (i = h-1; ++i < n; ) {  /* traverse the eigenvalues */
      if (vecs) r = vecs[i];    /* note the value to insert and do */
      v = evs[d = i];           /* insertion sort with increment h */
      do {                      /* to find the insertion position */
        s = d-h;                /* traverse the sorted eigenvalues */
        if (dir > 0) { if (v >= evs[s]) break; }
        else         { if (v <= evs[s]) break; }
        evs[d] = evs[s];        /* shift up all eigenvalues that have */
        if (vecs) vecs[d] = vecs[s];   /* to follow the current value */
      } while ((d = s) >= h);   /* while another shift is possible */
      evs[d] = v;               /* store the eigenvalue and */
      if (vecs) vecs[d] = r;    /* the corresponding eigenvector */
    }                           /* at the insertion position */
  }
  if (mat) mat_transp(mat,mat); /* turn eigenvectors into columns */
}  /* mat_evsort() */

/*----------------------------------------------------------------------
The eigenvalues are sorted with Shellsort, as it is easy to program and
still reasonably fast for commonly occurring matrix sizes. Compared to
the time needed to find the eigenvalues, even the time to transpose the
orthogonal matrix in the function above, the time is negligible anyway.
----------------------------------------------------------------------*/

void mat_bal (MATRIX *res, const MATRIX *mat)
{
}  /* mat_bal() */

/*--------------------------------------------------------------------*/

void mat_heselm (MATRIX *res, const MATRIX *mat)
{
}  /* mat_heselm() */

/*--------------------------------------------------------------------*/

void mat_hesqr (double *real, double *img, MATRIX *mat, int mode)
{
}  /* mat_hesqr() */
